;*********************************************************************
; BOXOTRIX ~ MIDI-To-Gate Converter                                  *
;*********************************************************************
; (c) 2006 SM-Sound / Seb Francis                                    *
;*********************************************************************
;
; Main code
;
;*********************************************************************
;
;    Connections:
;
;    RA0   <- !MIDI Channel bit0
;    RA1   <- !MIDI Channel bit1
;    RA2   <- !MIDI Channel bit2
;    RA3   <- !MIDI Channel bit3
;    RA4   n/c
;    RA5   n/c
;
;    RC0   n/c
;    RC1   n/c
;    RC2   -> GATE 1 (even MIDI notes)
;    RC3   -> GATE 2 (odd MIDI notes)
;    RC4   n/c
;    RC5   <- MIDI IN
;
;
;*********************************************************************


	list      p=16f688	; List directive to define processor
	#include <P16F688.inc>
	#include <const.inc>	; Constants + Memory map
	#include <misc.inc>	; Various macros
	
	
	__CONFIG _FCMEN_OFF & _IESO_OFF & _BOD_OFF & _CPD_OFF & _CP_OFF & _MCLRE_OFF & _PWRTE_ON & _WDT_OFF & _INTOSCIO

; '__CONFIG' directive is used to embed configuration data within .asm file.
; The lables following the directive are located in the respective .inc file.
; See respective data sheet for additional information on configuration word.




;**********************************************************************


	org     0x0000		; Processor reset vector
	clrf    PCLATH		; Ensure page bits are cleared
	goto    Init		; Go to beginning of program


	org     0x0004		; Interrupt vector location
	movwf   W_TEMP		; Save W, STATUS, PCLATH and FSR
	swapf	STATUS,w	; Use swapf, not movf so not to disturb Z flag
	movwf	STATUS_TEMP
	movfw   PCLATH
	movwf	PCLATH_TEMP
	movfw	FSR
	movwf	FSR_TEMP
	
	clrf    PCLATH		; Clear page bits
	

; Call the interupt handlers

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	PIR1,RCIF
	call	MIDIByteReceived

; --
	movfw	FSR_TEMP	; Restore W, STATUS, PCLATH and FSR
	movwf	FSR
	movfw	PCLATH_TEMP
	movwf	PCLATH
	swapf	STATUS_TEMP,w	; Use swapf, not movf so not to disturb Z flag
	movwf	STATUS
	swapf	W_TEMP,f	; Use swapf, not movf so not to disturb Z flag
	swapf	W_TEMP,w	; Swap nibbles again to get them correct
	retfie			; Return from interrupt


;**********************************************************************


Init

;
; Setup the I/O and SFR
; 

; Bank 0 Setup
	bcf	STATUS,RP0
	bcf	STATUS,RP1
	
	clrf	PORTC           ; GATE = 0
	
	movlw	b'01000000'	; Interupts: Global off, Peripheral on, TMR0 off, External off, PORTA change off
	movwf	INTCON
	
	clrf	PIR1		; Clear all interrupt flags
	
	clrf	T1CON		; Timer1 off
	
	; USART config
	clrf	BAUDCTL		; 8-bit baud rate, Wake up interrupts off, Auto Baud Detect off
	clrf	SPBRGH
	movlw	d'1'		; Set baud rate to 31250 ~ 4,000,000 / (64 * (1 + 1))
	movwf	SPBRG
	clrf	TXSTA		; Transmit disbaled, USART in Asynchronous mode, Low baud rate
	movlw	b'10000000'	; Serial port enabled, 8-bit reception, CREN=0 for now, Address Detect off
	movwf	RCSTA
	
	clrf	WDTCON		; Watchdog timer off
	movlw	b'00000111'	; Comparators off
	movwf	CMCON0
	clrf	ADCON0		; A/D Converter off
	
	
	; Initialise general data memory
	
	clrf	MIDI_FLAGS	; Clear MIDI flags
	clrf	MIDI_MSG1	; Clear MIDI status buffer
	
	movlw	low MIDI_MSG_BUFFER	; Set pointers to MIDI ring buffer
	movwf	MIDI_MSG_PROCESS_PTR
	movwf	MIDI_MSG_INCOMING_PTR
	
	
; Bank 1 Setup
	bsf	STATUS,RP0
	bcf	STATUS,RP1
	
	clrf	OPTION_REG	; PORTA pullups enabled, Internal TMR0 clock, Pre-scaler assigned to TMR0, Pre-scaler ratio 1:2
	
	movlw	b'00001111'	; Port A (RA0-RA3 inputs)
	movwf	TRISA
	movwf	WPUA		; Enable weak pull-ups for RA0-RA2
	clrf	IOCA		; Disable interrupt on change
	
	movlw	b'00100000'	; Port C (RC5 input)
	movwf	TRISC
	
	movlw	b'00100000'	; Peripheral interrupts: USART on, everything else off
	movwf	PIE1
	
	clrf	PCON		; Power Control Register
	
	movlw	b'01100001'	; Clock source is internal 4MHz oscillator
	movwf	OSCCON
	clrf	OSCTUNE		; Calibrated internal clock frequency
	
	clrf	ANSEL		; All pins digital I/O
	
	clrf	VRCON		; Voltage reference off


; Bank 2 Setup
; 	bcf	STATUS,RP0
; 	bsf	STATUS,RP1
	

; Bank 3 Setup
; 	bsf	STATUS,RP0
; 	bsf	STATUS,RP1
	

; --------------------

	clrf	STATUS		; Bank 0 Select
	
	bsf	RCSTA,CREN	; Enable MIDI receive
	bsf	INTCON,GIE	; Enable interrupts


;#######################################################################

; RAM Bank 0 !
Main

	movfw	MIDI_MSG_PROCESS_PTR	; Check if buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	jz	Main
	
	movlw	high ProcessMIDI
	movwf	PCLATH			; Select program bank 1, and ready PCLATH for jump table
	call	ProcessMIDI
	clrf	PCLATH
	goto	Main
	

;#######################################################################



; --------------------------------


PointToNextIncomingBufferByte	macro
	incf	FSR			; Increment pointer
	btfss	FSR,4			; Check not wrapped
	goto	$+3
	movlw	low MIDI_MSG_BUFFER	; Wrap pointer
	movwf	FSR
	
	movfw	FSR			; Check for ring buffer overrun
	subwf	MIDI_MSG_PROCESS_PTR,w
	jz	MIDIByteReceived_RingBuffOver
	endm

MIDIByteReceived_RingBuffOver
	clrf	MIDI_MSG1		; Clear status byte and flags
	clrf	MIDI_FLAGS
	return


; RAM bank 0
MIDIByteReceived

	movfw	RCSTA			; Read and save and error code
	movwf	TEMP2
	
	movfw	RCREG			; Read MIDI byte
	movwf	TEMP1			; Save in TEMP1
	
	btfsc	TEMP2,OERR		; Check for overrun error
	goto	MIDIByteReceived_Error
	btfsc	TEMP2,FERR		; Check for framing error
	goto	MIDIByteReceived_Error
	
	
	btfss	TEMP1,7			; Header or data byte?
	goto	MIDIByteReceived_DataByte
	
; header byte
	movlw	b'11110000'		; System message
	andwf	TEMP1,w
	sublw	b'11110000'
	jz	MIDIByteReceived_System
	
	; Voice message
	clrf	MIDI_FLAGS		; Clear flags
	
	movfw	TEMP1			; Store status byte
	movwf	MIDI_MSG1
	
	movlw	b'11100000'		; 2 byte messages (Program Change + Channel Pressure)
	andwf	TEMP1,w
	sublw	b'11000000'
	sz				; If not 2 byte message, set flag
	bsf	MIDI_FLAGS,MSG_LEN
	
	return
	
MIDIByteReceived_System
	btfss	TEMP1,3			; System common message?
	goto	MIDIByteReceived_SystemCommon
	
	; System realtime message
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write byte
	movwf	INDF
	
	return
	
MIDIByteReceived_SystemCommon
	clrf	MIDI_MSG1		; Clear status buffer
	
	return				; Ignore system common messages
	

MIDIByteReceived_DataByte
	movfw	MIDI_MSG1		; Drop byte if status is 0
	rz
	
	btfsc	MIDI_FLAGS,MSG_LEN	; Test if expecting 3 byte message
	goto	MIDIByteReceived_DataByte_3byte
	
	; 2 byte message
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG1		; Write status byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write data byte
	movwf	INDF
	
	return
	
MIDIByteReceived_DataByte_3byte
	btfsc	MIDI_FLAGS,WHICH_BYTE	; Test if this is the 2nd data byte
	goto	MIDIByteReceived_DataByte_3byte2
	
	; 3 byte message, 1st data byte
	movfw	TEMP1			; Store data byte
	movwf	MIDI_MSG2
	
	bsf	MIDI_FLAGS,WHICH_BYTE	; Set flag to expect 2nd data byte
	
	return
	
MIDIByteReceived_DataByte_3byte2
	; 3 byte message, 2nd data byte
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG1		; Write status byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG2		; Write 1st data byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write 2nd data byte
	movwf	INDF
	
	bcf	MIDI_FLAGS,WHICH_BYTE	; Set flag to expect 1st data byte
	
	return
	
	
MIDIByteReceived_Error
	bcf	RCSTA,CREN		; Reset receive logic / clear error flag
	bsf	RCSTA,CREN
	return
	
	
	
;#######################################################################

; More include files can go here ..
	
;#######################################################################


; Program bank 1

	org	0x0800
	
	; Implemented messages:
	;
	; 1001cccc nnnnnnnn vvvvvvvv  Note On
	; 1000cccc nnnnnnnn vvvvvvvv  Note Off


PointToNextProcessBufferByte	macro
	incf	MIDI_MSG_PROCESS_PTR	; Increment pointer
	btfss	MIDI_MSG_PROCESS_PTR,4	; Check not wrapped
	goto	$+3
	movlw	low MIDI_MSG_BUFFER	; Wrap pointer
	movwf	MIDI_MSG_PROCESS_PTR
	endm


; Parse MIDI message here

ProcessMIDI
; RAM Bank 0
	PointToNextProcessBufferByte
	
	bcf	STATUS,IRP	; Indirectly address banks 0-1
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	movfw	INDF		; Get and save MIDI channel
	andlw	b'00001111'
	movwf	MTEMP1
	
	rrf	INDF,w		; Get message
	movwf	MTEMP2
	rrf	MTEMP2
	rrf	MTEMP2
	rrf	MTEMP2,w
	andlw	b'00000111'
	
	addwf	PCL		; Jump to handler (PCLATH already set before call to ProcessMIDI)
	goto	ProcessMIDI_NoteOff
	goto	ProcessMIDI_NoteOn
	goto	ProcessMIDI_PolyAftertouch
	goto	ProcessMIDI_ControlChange
	goto	ProcessMIDI_ProgramChange
	goto	ProcessMIDI_Aftertouch
	goto	ProcessMIDI_PitchBend
	goto	ProcessMIDI_System


;***********************************************************************
;***********************************************************************
	
; Message handlers

ProcessMIDI_NoteOff
; RAM Bank 0
	; Get note value to MTEMP2
	PointToNextProcessBufferByte
 	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	movfw	INDF
	movwf	MTEMP2
	
	; Ignore velocity value
	PointToNextProcessBufferByte
	
ProcessMIDI_NoteOff_2
	; Check MIDI channel
	comf	PORTA,w
	andlw	B'00001111'
	subwf	MTEMP1,w
	rnz
	
	; If even MIDI note, set Gate 1 off
	btfss	MTEMP2,0
	bcf	PORTC,GATE1
	
	; If odd MIDI note, set Gate 2 off
	btfsc	MTEMP2,0
	bcf	PORTC,GATE2
	
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_NoteOn
; RAM Bank 0
	; Get note value to MTEMP2
	PointToNextProcessBufferByte
 	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	movfw	INDF
	movwf	MTEMP2

	; Get velocity value
	PointToNextProcessBufferByte
 	movfw	MIDI_MSG_PROCESS_PTR
 	movwf	FSR
 	movf	INDF,f
 	jz	ProcessMIDI_NoteOff_2	; If velocity=0 jump to note off code
	
	; Check MIDI channel
	comf	PORTA,w
	andlw	B'00001111'
	subwf	MTEMP1,w
	rnz
	
	; If even MIDI note, set Gate 1 on
	btfss	MTEMP2,0
	bsf	PORTC,GATE1
	
	; If odd MIDI note, set Gate 2 on
	btfsc	MTEMP2,0
	bsf	PORTC,GATE2
	
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_PolyAftertouch
	PointToNextProcessBufferByte
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_ControlChange
	; Point to controller number
	PointToNextProcessBufferByte
	
	; Check MIDI channel
	comf	PORTA,w
	andlw	B'00001111'
	subwf	MTEMP1,w
	jnz	ProcessMIDI_ControlChange_End
	
	; Get controller number
 	movfw	MIDI_MSG_PROCESS_PTR
 	movwf	FSR
 	movfw	INDF
 	
 	; All Sound Off message?
 	sublw	0x78
 	jz	ProcessMIDI_ControlChange_SndOff
 	
 	; All Notes Off message?
 	addlw	3
 	jz	ProcessMIDI_ControlChange_SndOff

ProcessMIDI_ControlChange_End
	; Ignore controller value
	PointToNextProcessBufferByte
	return
	
ProcessMIDI_ControlChange_SndOff
	; Turn off gates
	bcf	PORTC,GATE1
	bcf	PORTC,GATE2
	goto	ProcessMIDI_ControlChange_End
	
	
;***********************************************************************
	
	
ProcessMIDI_ProgramChange
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_Aftertouch
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_PitchBend
	PointToNextProcessBufferByte
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_System
	return
	
	
;***********************************************************************
;***********************************************************************
	
	

;#######################################################################


	END                       ; directive 'end of program'


